Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Open_on_rhs.Conv_failureSourceto_string_hum returns the JSON context stack as a string that reads well if printed directly. That string cannot be used as a sexp atom: the backslash-escaping would render it illegible.